from __future__ import annotations

import asyncio
import datetime
import functools

from typing import Any, Callable, Coroutine, TypeVar

import discord
from discord import ui, app_commands

from lgrez import commons
from lgrez.blocs import tools


class _DiscordJourneyView(ui.View):
    def __init__(self, journey: DiscordJourney, timeout: float | None = None):
        self.journey = journey
        super().__init__(timeout=timeout)

    async def interaction_check(self, interaction: discord.Interaction) -> bool:
        return await self.journey._interaction_check(interaction)

    async def on_error(self, _interaction: discord.Interaction, error: Exception, _item: ui.Item) -> None:
        raise error


class _DiscordJourneyModal(ui.Modal):
    def __init__(self, journey: DiscordJourney, title: str, timeout: float | None = None):
        self.journey = journey
        super().__init__(title=title, timeout=timeout)

    async def on_submit(self, interaction: discord.Interaction) -> None:
        self.journey.interaction = interaction
        self.stop()

    async def interaction_check(self, interaction: discord.Interaction) -> bool:
        return await self.journey._interaction_check(interaction)

    async def on_error(self, _interaction: discord.Interaction, error: Exception, _item: ui.Item) -> None:
        raise error


def _create_button(input: ui.Button | str | discord.Emoji) -> ui.Button:
    if isinstance(input, ui.Button):
        if input.label:
            input.label = input.label[:80]
        return input
    if isinstance(input, discord.Emoji):
        return ui.Button(emoji=input)
    if len(input) <= 2 and not input.isascii():
        return ui.Button(emoji=input)
    else:
        return ui.Button(label=str(input)[:80])


def _create_select_option(input: discord.SelectOption | str) -> discord.SelectOption:
    if isinstance(input, discord.SelectOption):
        return input
    return discord.SelectOption(label=input)


_Key = TypeVar("_Key")


class DiscordJourney:
    """Home-made protocol to make use of interactions easy when writing application commands.

    Args:
        interaction: The interaction that triggered the command.
        ephemeral: If ``True``, all messages are not public and shown only to the interaction triggerer.
        author_only: The only member allowed to interact ; if not passed, everyone can interact.
        timeout: The delay after which the interaction expires, in seconds.
    """

    RUNNING_INTERACTION: discord.Interaction = None
    TIMEOUT_DELAY: float = 2.0

    _catch_next_command_channel: discord.TextChannel | None = None
    _catch_next_command_member: discord.Member | None = None
    _catch_next_command_coro: Coroutine
    _tasks = set()

    def __init__(
        self,
        interaction: discord.Interaction,
        *,
        ephemeral: bool = False,
        author_only: discord.Member | None = None,
        timeout: float | None = None,
        command_author: discord.Member | None = None,
    ) -> None:
        self._thinking = True

        self.interaction = interaction
        self.ephemeral = ephemeral
        self.author_only = author_only
        self.timeout = timeout

        self.member = interaction.user
        if not isinstance(self.member, discord.Member):
            raise ValueError(f"Interaction user is not a guild member (type {type(self.member)})")
        self.command_author = command_author or self.member

        self.channel: discord.TextChannel = interaction.channel
        self.created_at: datetime.datetime = interaction.created_at

        self._stopped: bool = False

    @property
    def interaction(self) -> discord.Interaction:
        """The active interaction of this journey."""
        return self._interaction

    @interaction.setter
    def interaction(self, new_interaction: discord.Interaction) -> None:
        if not isinstance(new_interaction, discord.Interaction):
            raise TypeError(f"DiscordJourney.interaction must be a discord.Interaction! (got {type(new_interaction)})")
        self._interaction = new_interaction
        if self._thinking:
            asyncio.get_running_loop().call_later(self.TIMEOUT_DELAY, self._timeout_interaction, self.interaction)

    def _timeout_interaction(self, interaction: discord.Interaction) -> None:
        async def _timeout():
            if not interaction.response.is_done():
                await interaction.response.defer(thinking=True, ephemeral=self.ephemeral)

        task = asyncio.create_task(_timeout())
        self._tasks.add(task)  # Cf. https://docs.python.org/fr/3/library/asyncio-task.html#asyncio.create_task
        task.add_done_callback(self._tasks.discard)

    def last_interaction(self):
        self._thinking = False

    async def _send_message(
        self,
        content: str = "",
        view: ui.View | None = None,
        code: bool = False,
        prefix: str = "",
        ephemeral: bool = False,
        **kwargs,
    ) -> list[discord.Message]:
        return await tools.send_blocs(
            self.interaction,
            content,
            code=code,
            prefix=prefix,
            view=view,
            ephemeral=ephemeral or self.ephemeral,
            author=self.command_author,
            **kwargs,
        )

    async def send(self, content: str = "", code: bool = False, prefix: str = "", **kwargs) -> list[discord.Message]:
        """Send a message in this journey.

        Args:
            content: The textual contents to send.
            code: Whether to send the messages as code blocks.
            prefix: Is ``code`` is ``True``, optional text to send before the code blocks.
            kwargs: Passed to ``tools.send_blocs``.

        Returns:
            The messages posted.
        """
        return await self._send_message(content, code=code, prefix=prefix, **kwargs)

    def _check_state(self) -> None:
        if self._stopped:
            raise RuntimeError("Journey is stopped")

    async def _interaction_check(self, interaction: discord.Interaction) -> bool:
        if self.author_only and self.author_only != interaction.user:
            return False
        return True

    async def __aenter__(self):
        DiscordJourney.RUNNING_INTERACTION = None
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        if not self.interaction.response.is_done():
            if exc_type:
                DiscordJourney.RUNNING_INTERACTION = self.interaction
            else:
                try:
                    await self.interaction.response.defer()
                except discord.HTTPException:
                    pass
        self._stopped = True

    @staticmethod
    def _clicked_button_cleanup(button: ui.Button, view: ui.View) -> None:
        button.disabled = True

    @staticmethod
    def _not_clicked_button_cleanup(button: ui.Button, view: ui.View) -> None:
        view.remove_item(button)

    async def buttons(
        self,
        content: str,
        buttons: dict[_Key, ui.Button | str | discord.Emoji],
        clicked_button_cleanup: Callable[[ui.Button, ui.View], None] = _clicked_button_cleanup,
        not_clicked_button_cleanup: Callable[[ui.Button, ui.View], None] = _not_clicked_button_cleanup,
        **kwargs,
    ) -> _Key:
        """Send a message with buttons and wait for a click on one of them.

        Args:
            content: The message to send.
            buttons: The buttons to show, indexed by any hashable key, described either with a complete
                ``discord.ui.Button``, a simple label or an emoji.
            clicked_button_cleanup: What to do to the clicked button. Marks it disabled by default.
            not_clicked_button_cleanup: What to do to other buttons. Delete them by default.
            kwargs: Passed to ``journey.send``.

        Returns:
            The key index of the clicked button.
        """
        self._check_state()
        view = _DiscordJourneyView(self, timeout=self.timeout)
        clicked_button = None

        def _get_button_callback(button: ui.Button) -> Callable[[discord.Interaction], Coroutine]:
            async def _callback(interaction: discord.Interaction):
                nonlocal clicked_button
                self.interaction = interaction
                clicked_button = button
                view.stop()

            return _callback

        for key, button in buttons.items():
            button = _create_button(button)
            button.callback = _get_button_callback(button)
            view.add_item(button)
            buttons[key] = button

        *_, message = await self._send_message(content, view, **kwargs)  # Show buttons to user
        try:
            timeout = await view.wait()  # Wait for user click
        except asyncio.CancelledError:
            pass  # Clean view when cancelled
        else:
            if timeout:
                raise RuntimeError("Journey timeout")
        finally:
            # Clean view
            clicked_key = None
            for key, button in buttons.items():
                if button == clicked_button:
                    clicked_key = key
                    clicked_button_cleanup(button, view)
                else:
                    not_clicked_button_cleanup(button, view)
            await message.edit(view=view)

        return clicked_key

    async def yes_no(self, content: str, yes_emoji="✅", no_emoji="❌", **kwargs) -> bool:
        """Send a message with positive and negative buttons and wait for a click on one of them.

        Args:
            content: The message to send.
            yes_emoji: The emoji of the positive button.
            no_emoji: The emoji of the negative button.
            kwargs: Passed to ``journey.buttons``.

        Returns:
            The key index of the clicked button.
        """
        buttons = {
            True: ui.Button(emoji=yes_emoji, style=discord.ButtonStyle.success),
            False: ui.Button(emoji=no_emoji, style=discord.ButtonStyle.secondary),
        }
        return await self.buttons(content, buttons, **kwargs)

    async def ok_cancel(self, content: str, ok_emoji="✅", cancel_emoji="❌", **kwargs) -> None:
        """Send a message with positive and negative buttons, aborts command execution if the latter is clicked.

        Args:
            content: The message to send.
            ok_emoji: The emoji of the positive button.
            cancel_emoji: The emoji of the negative button.
            kwargs: Passed to ``journey.buttons``.

        Raises:
            commons.CommandAbortedError: If the cancel button is clicked.
        """
        ok_button = ui.Button(emoji=ok_emoji, style=discord.ButtonStyle.success)
        cancel_button = ui.Button(emoji=cancel_emoji, style=discord.ButtonStyle.secondary)

        def clicked_button_cleanup(button: ui.Button, view: ui.View) -> None:
            self._clicked_button_cleanup(button, view)
            if button == cancel_button:
                button.label = "Mission aborted."

        if not await self.buttons(
            content, {True: ok_button, False: cancel_button}, clicked_button_cleanup=clicked_button_cleanup, **kwargs
        ):
            await self.interaction.response.defer()
            raise commons.CommandAbortedError()

    async def select(
        self,
        content: str,
        options: dict[_Key, discord.SelectOption | str],
        placeholder: str | None = None,
        validate: str | None = None,
    ) -> _Key:
        """Send a message with a select menu and wait for one choice to be picked.

        Args:
            content: The message to send.
            options: The options of the select menu, indexed by any hashable key, described either with a
                complete ``discord.ui.SelectOption`` or a simple label.
            placeholder: Optional placeholder text, the first text is selected if not passed.
            validate: If passed, add a button next to the select menu with this label, to click on
                to validate the choice.

        Returns:
            The key index of the clicked button.
        """
        self._check_state()
        view = _DiscordJourneyView(self, timeout=self.timeout)

        options = {key: _create_select_option(option) for key, option in options.items()}
        select = ui.Select(options=options.values(), placeholder=placeholder)
        view.add_item(select)

        async def _callback(interaction: discord.Interaction):
            self.interaction = interaction
            view.stop()

        if validate:
            button = ui.Button(label=validate)
            view.add_item(button)
            button.callback = _callback

            async def _noop_callback(interaction: discord.Interaction):
                await interaction.response.defer()

            select.callback = _noop_callback
        else:
            select.callback = _callback

        *_, message = await self._send_message(content, view)  # Show select to user
        try:
            timeout = await view.wait()  # Wait for user click
        except asyncio.CancelledError:
            pass  # Clean view and exit
        else:
            if timeout:
                raise RuntimeError("Journey timeout")

        # Clean view
        if select.values:
            [value] = select.values
            key, option = next((key, option) for key, option in options.items() if option.value == value)
            option.default = True
            select.options = [option]
            select.disabled = True
        else:
            view.remove_item(select)
        if validate:
            view.remove_item(button)
        await message.edit(view=view)

        return key

    async def modal(self, title: str, *items: discord.ui.TextInput | str) -> list[str]:
        """Show a modal and wait for it's completion.

        Args:
            title: The modal title.
            items: The fields to include in the modal, described either with a complete
                ``discord.ui.TextInput`` or a simple label.

        Returns:
            The list of values filled in the fields, in the same order.
        """
        self._check_state()
        modal = _DiscordJourneyModal(self, title=title, timeout=self.timeout)

        inputs = [ui.TextInput(label=item) if isinstance(item, str) else item for item in items]
        for input in inputs:
            modal.add_item(input)

        if self.interaction.response.is_done():
            raise RuntimeError("No active interaction, cannot send modal")

        await self.interaction.response.send_modal(modal)  # Show modal to user
        timeout = await modal.wait()  # Wait for modal completion
        if timeout:
            raise RuntimeError("Journey timeout")
        self._check_state()

        return [input.value for input in inputs]

    async def command_parameters_modal(
        self, command: app_commands.Command, title_prefix: str = ""
    ) -> tuple[Callable[[DiscordJourney], None], dict[str, Any]]:
        """Open a modal allowing to chose parameters values for a command."""
        if command.parameters:
            values = await self.modal(
                f"{title_prefix} /{command.qualified_name}",
                *[
                    ui.TextInput(label=param.display_name, placeholder=param.description, required=param.required)
                    for param in command.parameters
                ],
            )
        else:
            values = []

        callback = command._callback._callable
        parameters = {
            parameter.name: await parameter._Parameter__parent._annotation.transform(self.interaction, value)
            for parameter, value in zip(command.parameters, values)
        }
        return callback, parameters

    async def _catch_next_command(self) -> tuple[discord.Interaction, Callable[[DiscordJourney], Coroutine]]:
        class _Awaitable:
            # Oh, well... let's say it just work?
            # If you really want to dig, begin with https://stackoverflow.com/a/60118660
            def __await__(self):
                while True:
                    result = yield
                    if result:
                        yield
                        return result

        async def _catch_next_command():
            return await _Awaitable()

        self._catch_next_command_channel = self.channel
        self._catch_next_command_member = self.member
        self._catch_next_command_coro = _catch_next_command()

        try:
            return await self._catch_next_command_coro
        finally:  # Stop catching
            self._catch_next_command_channel = None
            self._catch_next_command_member = None
            self._catch_next_command_coro = None

    async def catch_next_command(
        self, next_command_message: str
    ) -> tuple[discord.Interaction, Callable[[DiscordJourney], Coroutine]]:
        """[WIP - does not work?] Catch the next command in the channel (or abort it with a cancel button).

        Args:
            next_command_message: The message describing which command to send next.

        Raises:
            commons.CommandAbortedError: If the cancel button is clicked, or if an exception occurs.

        Returns:
            The interaction that triggered the catched command, and a coroutine function to run it.
        """
        catch_task = asyncio.Task(self._catch_next_command())
        cancel_task = asyncio.Task(
            self.buttons(
                f"{next_command_message} :arrow_down:",
                {"cancel": discord.ui.Button(label="Annuler", emoji="❌")},
                ephemeral=True,
            )
        )
        done, _ = await asyncio.wait([catch_task, cancel_task], return_when=asyncio.FIRST_COMPLETED)

        match list(done):
            case [task] if task == catch_task:
                cancel_task.cancel()  # Remove cancel button
                return catch_task.result()
            case [task] if task == cancel_task:
                catch_task.cancel()  # Stop command catching
                raise commons.CommandAbortedError()
            case []:  # Exception in one of the coroutines
                catch_task.cancel()
                cancel_task.cancel()
                raise commons.CommandAbortedError()
            case _:
                raise RuntimeError(f"heu -> {done}")

    @classmethod
    def _journey_command(cls, callable):
        @functools.wraps(callable)
        async def new_callable(interaction: discord.Interaction, **kwargs):
            async def _run_command(_journey: DiscordJourney):  # Closure
                await callable(_journey, **kwargs)

            if (
                cls._catch_next_command_channel
                and interaction.channel == cls._catch_next_command_channel
                and interaction.user == cls._catch_next_command_member
            ):
                cls._catch_next_command_channel = None
                cls._catch_next_command_member = None
                cls._catch_next_command_coro.send((interaction, _run_command))
                return  # Do not execute command

            async with cls(interaction) as journey:
                await _run_command(journey)

        new_callable._callable = callable

        return new_callable


def journey_command(callable: Callable[[DiscordJourney], None]) -> Callable[[discord.Interaction], None]:
    """Decorator: add the Journey interface to a slash command.

    Replaces the ``interaction: discord.Interaction`` in the callable signature by ``journey: DiscordJourney``.
    """
    return DiscordJourney._journey_command(callable)


_CT = TypeVar("_CT", discord.Message, discord.Member)


def journey_context_menu(callable: Callable[[DiscordJourney, _CT], None]) -> Callable[[discord.Interaction, _CT], None]:
    """Decorator: add the Journey interface to a context menu command.

    Replaces the ``interaction: discord.Interaction`` in the callable signature by ``journey: DiscordJourney``.
    """

    @functools.wraps(callable)
    async def new_callable(interaction: discord.Interaction, arg: _CT):
        async with DiscordJourney(interaction, ephemeral=True) as journey:
            await callable(journey, arg)

    return new_callable



"""lg-rez / features / Commandes annexes

Commandes diverses qu'on ne savait pas où ranger

"""

import random
import requests
import datetime

from discord import app_commands
from akinator.async_aki import Akinator

from lgrez import config, commons
from lgrez.blocs import tools
from lgrez.blocs.journey import DiscordJourney, journey_command
from lgrez.bdd import Joueur, Role, Camp


DESCRIPTION = """Commandes annexes aux usages divers"""

next_roll = None


@app_commands.command()
@journey_command
async def roll(journey: DiscordJourney, *, pattern: str):
    """Lance un ou plusieurs dés, ou tire un nom dans un liste.

    Args:
        pattern: Dés à lancer (XdY [+ ...] [+ Z ...]) OU roll spécial (joueur / vivant / mort / rôle / camp).

    Examples:
        - ``/roll 1d6``           -> lance un dé à 6 faces
        - ``/roll 1d20 +3``       -> lance un dé à 20 faces, ajoute 3 au résultat
        - ``/roll 1d20 + 2d6 -8`` -> lance un dé 20 plus deux dés 6, enlève 8 au résultat
        - ``/roll vivant``        -> choisit un joueur vivant
    """
    global next_roll
    moi = Joueur.from_member(journey.member)

    inp = pattern.lower()
    # Rolls spéciaux
    result = None
    if inp in ["joueur", "joueurs"]:
        result = random.choice(Joueur.query.all()).nom
    elif inp in ["vivant", "vivants"]:
        jr = random.choice(Joueur.query.filter(Joueur.est_vivant).all())
        result = jr.nom
    elif inp in ["mort", "morts"]:
        jr = random.choice(Joueur.query.filter(Joueur.est_mort).all())
        result = jr.nom
    elif inp in ["role", "rôle", "roles", "rôles"]:
        role = random.choice(Role.query.filter_by(actif=True).all())
        result = role.nom_complet
    elif inp in ["camp", "camps"]:
        result = random.choice(Camp.query.filter_by(public=True).all()).nom
    elif inp in ["ludo", "ludopathe"]:
        result = random.choice(["Voyante", "Marionnettiste", "Notaire", "Popo de mort", "Chat-garou", "Espion"])
    elif inp in ["taverne", "tavernier"]:
        result = random.choice(["Rôle choisi", "Vrai rôle", "Rôle random"])
    elif inp in ["loup", "méchants", "nécro", "necromants", "nécromancien", "nécromanciens", "loups", "meute"]:
    	result = moi

    if result:
        if next_roll is not None:
            result = next_roll
            next_roll = None
        await journey.send(result)
        return

    parts = inp.replace(" ", "").replace("-", "+-").split("+")
    # "1d6 + 5 - 2" -> ["1d6", "5", "-2"]
    sum = 0
    rep = ""
    for part in parts:
        if not part:
            continue
        if "d" in part:
            # Lancer de dé
            nb, _, faces = part.partition("d")
            try:
                nb, faces = int(nb), int(faces)
                if faces < 1:
                    raise ValueError
            except ValueError:
                raise commons.UserInputError("pattern", f"Pattern de dé non reconnu : {part}")
            # Sécurité
            if abs(nb) > 1000 or faces > 1000000:
                await journey.send.send(
                    "Suite à des abus (coucou Gabin), il est "
                    "interdit de lancer plus de 1000 dés ou "
                    "des dés à plus de 1 million de faces."
                )
                return
            sig = -1 if nb < 0 else 1
            sig_s = "-" if nb < 0 else "+"
            for _ in range(abs(nb)):
                val = random.randrange(faces) + 1
                if next_roll is not None:
                    try:
                        val = int(next_roll)
                    except ValueError:
                        pass
                    else:
                        next_roll = None
                sum += sig * val
                rep += f" {sig_s} {val}₍{tools.sub_chiffre(faces, True)}₎"
        else:
            # Bonus / malus fixe
            try:
                val = int(part)
            except ValueError:
                raise commons.UserInputError("pattern", f"Pattern fixe non reconnu : {part}")
            sum += val
            rep += f" {'-' if val < 0 else '+'} {abs(val)}"
    # Total
    sig = "- " if sum < 0 else ""
    rep += f" = {sig}{tools.emoji_chiffre(abs(sum), True)}"
    rep = rep[3:] if rep.startswith(" +") else rep
    await journey.send(rep)


@app_commands.command()
@tools.mjs_only
@journey_command
async def nextroll(journey: DiscordJourney, *, next: str = None):
    """✨ Shhhhhhhhhhhh.

    Ça sent la magouilleuh
    """
    global next_roll
    next_roll = next
    await journey.send("🤫", ephemeral=True)


@journey_command
@app_commands.command()
async def coinflip(journey: DiscordJourney):
    """Pile ou face ?

    Pile je gagne, face tu perds.
    """
    await journey.send(random.choice(["Pile", "Face"]))


@app_commands.command()
@journey_command
async def ping(journey: DiscordJourney):
    """Envoie un ping au bot, pour vérifier si il est réactif tout ça.

    Pong
    """
    ts_rec = datetime.datetime.now(datetime.timezone.utc)
    delta_rec = ts_rec - journey.created_at
    # Temps de réception = temps entre création message et sa réception

    cont = f" Réception : {delta_rec.total_seconds()*1000:4.0f} ms\n Latence :   {config.bot.latency*1000:4.0f} ms\n"
    *_, mess = await journey.send(f"pong\n" + tools.code_bloc(cont + " (...)"))

    ts_ret = datetime.datetime.now(datetime.timezone.utc)
    delta_ret = ts_ret - mess.created_at
    # Retour information message réponse créé
    delta_env = ts_ret - ts_rec - delta_ret
    # Temps d'envoi = temps entre réception 1er message (traitement quasi instantané)
    # et création 2e, moins temps de retour d'information
    delta_tot = delta_rec + delta_ret + delta_env
    # Total = temps entre création message /pong et réception information réponse envoyée
    await mess.edit(
        content=f"/pong\n"
        + tools.code_bloc(
            cont + f" Envoi :     {delta_env.total_seconds()*1000:4.0f} ms\n"
            f" Retour :    {delta_ret.total_seconds()*1000:4.0f} ms\n"
            f"——————————————————————\n"
            f" Total :     {delta_tot.total_seconds()*1000:4.0f} ms"
        )
    )


@app_commands.command()
@journey_command
async def akinator(journey: DiscordJourney):
    """J'ai glissé chef

    Implémentation directe de https://pypi.org/project/akinator.py
    """
    # Un jour mettre ça dans des embeds avec les https://fr.akinator.com/
    # bundles/elokencesite/images/akitudes_670x1096/<akitude>.png croppées,
    # <akitude> in ["defi", "serein", "inspiration_legere",
    # "inspiration_forte", "confiant", "mobile", "leger_decouragement",
    # "vrai_decouragement", "deception", "triomphe"]
    await journey.send(
        "Vous avez demandé à être mis en relation avec "
        + tools.ital("Akinator : Le Génie du web")
        + ".\nVeuillez patienter..."
    )

    chan = journey.channel
    async with chan.typing():
        # Connexion
        aki = Akinator()
        question = await aki.start_game(language="fr")

    while aki.progression <= 80:
        reponse = await journey.buttons(
            f"({aki.step + 1}) {question}", {"yes": "👍", "idk": "🤷", "no": "👎", "stop": "⏭️"}
        )
        if reponse == "stop":
            break
        question = await aki.answer(reponse)

    await aki.win()

    if await journey.yes_no(
        f"Tu penses à {tools.bold(aki.first_guess['name'])} "
        f"({tools.ital(aki.first_guess['description'])}) !\n"
        f"J'ai bon ?\n{aki.first_guess['absolute_picture_path']}"
    ):
        await chan.send("Yay\nhttps://fr.akinator.com/bundles/elokencesite/images/akitudes_670x1096/triomphe.png")
    else:
        await chan.send("Oof\nhttps://fr.akinator.com/bundles/elokencesite/images/akitudes_670x1096/deception.png")


@app_commands.command()
@journey_command
async def xkcd(journey: DiscordJourney, num: int):
    """J'ai aussi glissé chef, mais un peu moins

    Args:
        num: Numéro de la planche à récupérer.
    """
    rep = requests.get(f"https://xkcd.com/{num}/info.0.json")

    if not rep:
        await journey.send(":x: Paramètre incorrect ou service non accessible.")
        return

    url = rep.json().get("img")
    if not url:
        await journey.send(":x: Paramètre incorrect ou service non accessible.")
        return

    await journey.send(url)
    



"""lg-rez / features / Gestion des actions

Liste, création, suppression, ouverture, fermeture d'actions

"""

import datetime
from typing import Literal

from lgrez import config, commons
from lgrez.blocs import tools
from lgrez.bdd import Action, BaseAction, Tache, Utilisation, ActionTrigger, Joueur


def add_action(
    joueur: Joueur, base: BaseAction, cooldown: int = 0, charges: int | None = None, active: bool = True
) -> Action:
    """Enregistre une action et programme son ouverture le cas échéant.

    Si une action existe déjà pour ce joueur et cette base, la modifie ; sinon, en crée une nouvelle.

    Args:
        joueur, base, cooldown, charges, active: Paramètres de l'action.
    """
    action = Action.query.filter_by(joueur=joueur, base=base).first()
    if action:
        action.cooldown = cooldown
        action.charges = charges
        action.active = active
        action.update()
    else:
        action = Action(joueur=joueur, base=base, cooldown=cooldown, charges=charges, active=active)
        action.add()

    # Ajout tâche ouverture
    if action.base.trigger_debut == ActionTrigger.temporel:
        # Temporel : on programme
        Tache(
            timestamp=tools.next_occurrence(action.base.heure_debut),
            commande="open action",
            parameters={"id": action.id},
            action=action,
        ).add()

    elif action.base.trigger_debut == ActionTrigger.perma:
        # Perma : ON LANCE DIRECT (sera repoussé si jeu fermé)
        Tache(
            timestamp=datetime.datetime.now(), commande="open action", parameters={"id": action.id}, action=action
        ).add()

    return action


def delete_action(action: Action) -> None:
    """Archive une action et annule les tâches en cours liées.

    Depuis la version 2.1, l'action n'est plus supprimée mais est passée à :attr:`~.bdd.Action.active` = ``False``.

    Args:
        action: L'action à supprimer.
    """
    # Suppression tâches liées à l'action
    if action.taches:
        Tache.delete(*action.taches)

    if action.is_open:
        action.utilisation_ouverte.close()

    action.active = False
    action.update()
    # action.delete()


async def open_action(action: Action) -> None:
    """Ouvre l'action.

    Args:
        action: L'action à ouvrir.

    Opérations réalisées :
        - Vérification des conditions (cooldown, charges...) et reprogrammation si nécessaire ;
        - Gestion des tâches planifiées (planifie remind/close si applicable) ;
        - Information du joueur.
    """
    joueur = action.joueur
    chan = joueur.private_chan

    # Vérification base
    if not action.base:
        await tools.log(f"{action} : pas de base, exit")
        return

    # Vérification active
    if not action.active:
        await tools.log(f"{action} : inactive, exit (pas de reprogrammation)")
        return

    # Vérification cooldown
    if action.cooldown > 0:  # Action en cooldown
        action.cooldown = action.cooldown - 1
        config.session.commit()
        await tools.log(f"{action} : en cooldown, exit (reprogrammation si temporel).")
        if action.base.trigger_debut == ActionTrigger.temporel:
            # Programmation action du lendemain
            ts = tools.next_occurrence(action.base.heure_debut)
            Tache(timestamp=ts, commande="open action", parameters={"id": action.id}, action=action).add()
        return

    # Vérification role_actif
    if not joueur.role_actif:
        # role_actif == False : on reprogramme la tâche au lendemain tanpis
        await tools.log(f"{action} : role_actif == False, exit (reprogrammation si temporel).")
        if action.base.trigger_debut == ActionTrigger.temporel:
            ts = tools.next_occurrence(action.base.heure_debut)
            Tache(timestamp=ts, commande="open action", parameters={"id": action.id}, action=action).add()
        return

    # Vérification charges
    if action.charges == 0:
        # Plus de charges, mais action maintenue en base car refill / ...
        await tools.log(f"{action} : plus de charges, exit (reprogrammation si temporel).")
        return

    # Action "automatiques" (passives : notaire...) :
    # lance la procédure de clôture / résolution
    if action.base.trigger_fin == ActionTrigger.auto:
        if action.base.trigger_debut == ActionTrigger.temporel:
            await tools.log(
                f"Action {action.base.slug} pour {joueur.nom} pas vraiment automatique, {config.Role.mj.mention} "
                f"VENEZ M'AIDER JE PANIQUE 😱 (comme je suis vraiment sympa je vous file son chan, {chan.mention})"
            )
        else:
            await tools.log(f"{action} : automatique, appel processus de clôture")

        await close_action(action)
        return

    # Tous tests préliminaires n'ont pas return ==> Vraie action à lancer

    # Calcul heure de fin (si applicable)
    heure_fin = None
    if action.base.trigger_fin == ActionTrigger.temporel:
        heure_fin = action.base.heure_fin
        ts = tools.next_occurrence(heure_fin)
    elif action.base.trigger_fin == ActionTrigger.delta:
        # Si delta, on calcule la vraie heure de fin (pas modifié en base)
        delta = action.base.heure_fin
        ts = datetime.datetime.now() + datetime.timedelta(hours=delta.hour, minutes=delta.minute, seconds=delta.second)
        heure_fin = ts.time()

    # Programmation remind / close
    if action.base.trigger_fin in [ActionTrigger.temporel, ActionTrigger.delta]:
        remind_ts = ts - datetime.timedelta(minutes=30)
        Tache(timestamp=remind_ts, commande="remind action", parameters={"id": action.id}, action=action).add()
        Tache(timestamp=ts, commande="close action", parameters={"id": action.id}, action=action).add()
    elif action.base.trigger_fin == ActionTrigger.perma:
        # Action permanente : fermer pour le WE
        # ou rappel / réinitialisation chaque jour
        ts_matin = tools.next_occurrence(datetime.time(hour=7))
        ts_pause = tools.debut_pause()
        if ts_matin < ts_pause:
            # Réopen le lendamain
            Tache(timestamp=ts_matin, commande="open action", parameters={"id": action.id}, action=action).add()
        else:
            # Sauf si pause d'ici là
            Tache(timestamp=ts_pause, commande="close action", parameters={"id": action.id}, action=action).add()

    # Information du joueur
    if action.is_open:  # déjà ouverte
        await chan.send(
            f"{tools.montre()}  Rappel : tu peux utiliser quand tu le souhaites "
            f"ton action {tools.code(action.base.slug)} !  {config.Emoji.action}\n"
            + (f"Tu as jusqu'à {heure_fin} pour le faire. \n" if heure_fin else "")
            + tools.ital(f"Utilise `/action` pour agir.")
        )
    else:
        # on ouvre !
        util = Utilisation(action=action)
        util.add()
        util.open()
        await chan.send(
            f"{tools.montre()}  Tu peux maintenant utiliser ton action "
            f"{tools.code(action.base.slug)} !  {config.Emoji.action}\n"
            + (f"Tu as jusqu'à {heure_fin} pour le faire. \n" if heure_fin else "")
            + tools.ital(f"Utilise `/action` pour agir.")
        )

    config.session.commit()


async def close_action(action: Action) -> None:
    """Ferme l'action.

    Args:
        action: L'action à enregistrer.

    Opérations réalisées :
        - Archivage si nécessaire ;
        - Gestion des tâches planifiées (planifie prochaine ouverture si applicable) ;
        - Information du joueur (si charge-- seulement).
    """
    joueur = action.joueur
    chan = joueur.private_chan

    # Vérification base
    if not action.base:
        await tools.log(f"{action} : pas de base, exit")
        return

    # Vérification active
    if not action.active:
        await tools.log(f"{action} : inactive, exit")
        return

    # Vérification ouverte
    if not action.is_open:
        await tools.log(f"{action} : pas ouverte, exit")
        return

    deleted = False
    if not action.is_waiting:  # décision prise
        # Résolution de l'action
        # (pour l'instant juste charge -= 1 et suppression le cas échéant)
        if action.charges:
            action.charges = action.charges - 1
            pcs = " pour cette semaine" if "weekends" in action.base.refill else ""
            await chan.send(f"Il te reste {action.charges} charge(s){pcs}.")

            if action.charges == 0 and not action.base.refill:
                delete_action(action)
                deleted = True

    if not deleted:
        # Si l'action a été faite et a un cooldown, on le met
        if (not action.is_waiting) and (action.base.base_cooldown > 0):
            action.cooldown = action.base.base_cooldown

        action.utilisation_ouverte.close()

        # Programmation prochaine ouverture
        if action.base.trigger_debut == ActionTrigger.temporel:
            ts = tools.next_occurrence(action.base.heure_debut)
            Tache(timestamp=ts, commande="open action", parameters={"id": action.id}, action=action).add()
        elif action.base.trigger_debut == ActionTrigger.perma:
            # Action permanente : ouvrir après le WE
            ts = tools.fin_pause()
            Tache(timestamp=ts, commande="open action", parameters={"id": action.id}, action=action).add()

    config.session.commit()


def get_actions(
    quoi: Literal["open", "close", "remind"], trigger: ActionTrigger, heure: datetime.time | None = None
) -> list[Action]:
    """Renvoie les actions répondant à un déclencheur donné.

    Args:
        quoi: Type d'opération en cours :

          - ``"open"`` : ouverture, :attr:`Action.is_open` doit être faux;
          - ``"close"`` :  fermeture, :attr:`Action.is_open` doit être vrai;
          - ``"remind"`` : rappel, :attr:`Action.is_waiting` doit être vrai.

        trigger: valeur de ``Action.trigger_debut/fin`` à détecter.
        heure: si ``trigger == "temporel"``, ajoute la condition ``Action.heure_debut/fin == heure``.

    Returns:
        La liste des actions correspondantes.
    """
    filter = Action.active.is_(True).join(Action.base)

    if quoi == "open":
        filter &= (BaseAction.trigger_debut == trigger) & ~Action.is_open
    elif quoi == "close":
        filter &= (BaseAction.trigger_fin == trigger) & Action.is_open
    elif quoi == "remind":
        filter &= (BaseAction.trigger_fin == trigger) & Action.is_waiting
    else:
        raise commons.UserInputError("quoi", f"bad value: '{quoi}'")

    if trigger == ActionTrigger.temporel:
        if not heure:
            raise commons.UserInputError("quand", "merci de préciser une heure")

        if quoi == "open":
            filter &= BaseAction.heure_debut == heure
        else:  # close / remind
            filter &= BaseAction.heure_fin == heure

    return Action.query.filter(filter).all()



"""lg-rez / blocs / Lecture des variables d'environnement

"""

import sys
import os

from dotenv import load_dotenv


# À l'import
if len(sys.argv) > 1:
    path = sys.argv[1]
else:
    path = ".env"

load_dotenv(path)


def load(var_name: str) -> str:
    """Lit une valeur depuis les variables d'environnement.

    Les variables sont recherchées dans les variables d'environnement ;
    à l'import de ce module, celles-ci sont chargées par
    ``dotenv.load_dotenv`` à partir du fichier passé en premier
    argument en ligne de commande (``.env`` par défaut).

    Équivaut globalement à :func:`os.getenv` suivi d'une vérification
    que la variable existe.

    Args:
        var_name: nom de la variable à charger (``LGREZ_...``).

    Returns:
        La valeur de la variable demandée.

    Raises:
        RuntimeError: la variable d'environnement n'est pas définie.
    """
    var = os.getenv(var_name)
    if var is None:
        raise RuntimeError(f"Variable d'environnement {var_name} manquante")
    return var


def __getattr__(attr: str) -> str:
    """Raccourci pour accéder aux variables d'environnement.

    Permet d'utiliser ``blocs.env.VAR_NAME`` comme raccourci pour
    :func:`blocs.env.load(VAR_NAME) <.load>`, à la différence près
    que c'est une exception :exc:`AttributeError` qui sera levée si
    la variable n'est pas définie.
    """
    try:
        return load(attr)
    except RuntimeError:
        raise AttributeError(f"module '{__name__}' has no attribute '{attr}'") from None



"""lg-rez / blocs / Outils divers et variés

Récupération d'objets Discord, décorateurs pour commandes, structures
d'interaction dans les channels, utilitaires d'emojis, de date/temps,
de formatage...

"""

from __future__ import annotations

import asyncio
import datetime
import re
import string
import typing
from typing import Any, Callable

import discord
import discord.utils
from discord import app_commands
import unidecode

from lgrez import config, bdd, commons
from lgrez.bdd import Camp, CandidHaro, CandidHaroType, Joueur, Role


# ---------------------------------------------------------------------------
# Utilitaires de récupération d'objets Discord (détectent les mentions)
# ---------------------------------------------------------------------------


def _get(collec, **attrs):
    for elem in collec:
        if all(getattr(elem, attr) == value for attr, value in attrs.items()):
            return elem
    return None


_T = typing.TypeVar("_T")


def _find_by_mention_or_name(
    collec: typing.Collection[_T],
    val: str,
    pattern: str | None = None,
    must_be_found: bool = False,
    raiser: str | None = None,
) -> _T | None:
    """Utilitaire pour la suite : trouve un élément dans une liste.

    Args:
        collection: liste d'éléments.
        val: nom de l'élément à trouver, correspondant à son attribut
            ``name`` ou respectant une regex.
        pattern: motif regex à utiliser pour la recherche, si besoin.
        must_be_found: Si ``True``, raise une :exc:`ValueError` si
            ``val`` est introuvable.
        raiser: nom de la fonction à envoyer dans l'exception si
            introuvable (défaut : cette fonction).

    Returns:
        L'élément recherché, si on l'a trouvé.

    Raises:
        ValueError: Si ``must_be_found`` est ``True`` et que l'élément
            n'a pas été trouvé.
    """
    if not val:
        item = None
    elif pattern and (match := re.search(pattern, val)):
        item = _get(collec, id=int(match.group(1)))
    else:
        item = _get(collec, name=val)

    if must_be_found and not item:
        if raiser is None:
            raiser = "tools._find_by_mention_or_name"
        raise ValueError(f"{raiser} : Élément '{val}' introuvable")

    return item


@typing.overload
def channel(nom: str, must_be_found: typing.Literal[True] = True) -> discord.TextChannel | discord.CategoryChannel:
    ...


@typing.overload
def channel(nom: str, must_be_found: typing.Literal[False]) -> discord.TextChannel | discord.CategoryChannel | None:
    ...


def channel(nom, must_be_found=True):
    """Renvoie l'objet associé au salon ``#nom``.

    Args:
        nom: nom du channel (texte/vocal/catégorie) ou sa
            mention (détection directe par regex)
        must_be_found: si ``True`` (défaut), raise une
            :exc:`ValueError` si le channel ``#nom`` n'existe pas
            (si ``False``, renvoie ``None``)

    Returns:
        Le salon ou la catégorie Discord correspondant.
    """
    return _find_by_mention_or_name(
        config.guild.channels, nom, pattern="<#([0-9]{18})>", must_be_found=must_be_found, raiser="tools.channel"
    )


@typing.overload
def role(nom: str, must_be_found: typing.Literal[True] = True) -> discord.Role:
    ...


@typing.overload
def role(nom: str, must_be_found: typing.Literal[False] = True) -> discord.Role | None:
    ...


def role(nom, must_be_found=True):
    """Renvoie l'objet associé au rôle ``@&nom``.

    Args:
        nom: nom du rôle ou sa mention
            (détection directe par regex)
        must_be_found: si ``True`` (défaut), raise une
            :exc:`ValueError` si le channel ``@&nom`` n'existe pas
            (si ``False``, renvoie ``None``)

    Returns:
        Le rôle Discord correspondant.
    """
    return _find_by_mention_or_name(
        config.guild.roles, nom, pattern="<@&([0-9]{18})>", must_be_found=must_be_found, raiser="tools.role"
    )


@typing.overload
def member(nom: str, must_be_found: typing.Literal[True] = True) -> discord.Member:
    ...


@typing.overload
def member(nom: str, must_be_found: typing.Literal[False] = True) -> discord.Member | None:
    ...


def member(nom, must_be_found=True):
    """Renvoie l'objet associé au membre ``@nom``.

    Args:
        nom: nom du joueur ou sa mention
            (détection directe par regex)
        must_be_found: si ``True`` (défaut),
            raise une :exc:`ValueError` si le membre ``@nom`` n'existe pas
            (si ``False``, renvoie ``None``)

    Returns:
        Le membre Discord correspondant.
    """
    return _find_by_mention_or_name(
        config.guild.members, nom, pattern="<@!([0-9]{18})>", must_be_found=must_be_found, raiser="tools.member"
    )


@typing.overload
def emoji(nom: str, must_be_found: typing.Literal[True] = True) -> discord.Emoji:
    ...


@typing.overload
def emoji(nom: str, must_be_found: typing.Literal[False] = True) -> discord.Emoji | None:
    ...


def emoji(nom, must_be_found=True):
    """Renvoie l'objet associé à l'emoji ``:nom:``.

    Args:
        nom: nom de l'emoji (texte/vocal/catégorie)
            ou son utilisation (détection directe par regex)
        must_be_found: si ``True`` (défaut), raise une
            :exc:`ValueError` si l'emoji ``:nom:`` n'existe pas
            (si ``False``, renvoie ``None``)

    Returns:
        L'emoji Discord correspondant.
    """
    return _find_by_mention_or_name(
        config.guild.emojis, nom, pattern="<:.*:([0-9]{18})>", must_be_found=must_be_found, raiser="tools.emoji"
    )


# Appel aux MJs
def mention_MJ(arg: discord.Member | discord.Interaction) -> str:
    """Renvoie la mention ou le nom du rôle MJ

        - Si le joueur n'est pas un MJ, renvoie la mention de
          :attr:`config.Role.mj`
        - Sinon, renvoie son nom (pour ne pas rameuter tout le monde).

    Args:
        arg: Le membre ou le contexte d'un message envoyé par un membre.

    Returns:
        La chaîne à utiliser pour mentionner le rôle MJ.
    """
    member = arg.user if isinstance(arg, discord.Interaction) else arg
    if config.is_setup and isinstance(member, discord.Member) and member.top_role == config.Role.mj:
        # Pas un webhook et (au moins) MJ
        return f"@{config.Role.mj.name}"
    else:
        return config.Role.mj.mention


# ---------------------------------------------------------------------------
# Décorateurs pour les différentes commandes, en fonction de leur usage
# ---------------------------------------------------------------------------


def mjs_only(callback):
    """Décorateur pour commande (:func:`discord.app_commands.check`)

    Commande exécutable uniquement par un :attr:`MJ <.config.Role.mj>`
    ou un webhook (tâche planifiée)
    """
    return app_commands.checks.has_role(config.Role.get_raw("mj"))(
        app_commands.default_permissions(manage_messages=True)(callback)
    )


def mjs_et_redacteurs(callback):
    """Décorateur pour commandes d'IA (:func:`discord.app_commands.check`)

    commande exécutable par un :attr:`MJ <.config.Role.mj>`, un
    :attr:`Rédacteur <.config.Role.redacteur>` ou un webhook (tâche planifiée)
    """
    return app_commands.checks.has_any_role(config.Role.get_raw("mj"), config.Role.get_raw("redacteur"))(
        app_commands.default_permissions(priority_speaker=True)(callback)  # Hack since we cannot restrict on roles
    )


def joueurs_only(callback):
    """Décorateur pour commande (:func:`discord.app_commands.check`) :

    commande exécutable uniquement par un joueur,
    :attr:`vivant <.config.Role.joueur_en_vie>` ou
    :attr:`mort <.config.Role.joueur_mort>`.
    """
    return app_commands.checks.has_any_role(config.Role.get_raw("joueur_en_vie"), config.Role.get_raw("joueur_mort"))(
        callback
    )


def vivants_only(callback):
    """Décorateur pour commande (:func:`discord.app_commands.check`) :

    commande exécutable uniquement par un
    :attr:`joueur vivant <.config.Role.joueur_en_vie>`
    """
    return app_commands.checks.has_role(config.Role.get_raw("joueur_en_vie"))(
        app_commands.default_permissions(send_messages=True)(callback)
    )


_Command = typing.TypeVar("_Command", bound=typing.Callable)


def private() -> Callable[[_Command], _Command]:
    """Décorateur : commande utilisable dans son chan privé uniquement.

    Lors d'une invocation de la commande décorée hors d'un channel privé
    (commençant par :attr:`config.private_chan_prefix`), supprime le
    message d'invocation et exécute la commande dans le channel privé de l'invocateur.

    Si le joueur ayant utilisé la commande n'a pas de chan privé (pas en base), raise une :exc:`ValueError`.

    Utilisable en combinaison avec :func:`.joueurs_only` et :func:`.vivants_only`
    (pas avec les autres attention, vu que seuls les joueurs ont un channel privé).
    """

    async def predicate(interaction: discord.Interaction) -> bool:
        if interaction.channel.name.startswith(config.private_chan_prefix):  # OK
            return True

        private_chan = Joueur.from_member(interaction.user).private_chan
        await interaction.response.send_message(
            f":x: La commande `/{interaction.command.qualified_name}` est interdite en dehors de ta conv privée !\n"
            f":arrow_forward: Réessaie dans {private_chan.mention}",
            ephemeral=True,
        )
        return False

    return app_commands.check(predicate)


# ---------------------------------------------------------------------------
# Transformers
# ---------------------------------------------------------------------------


_Table = typing.TypeVar("_Table", bound=bdd.base.TableBase)


class _TableTransformerMixin:
    async def _transform(self, table: type[_Table], value: str) -> _Table:
        elems = table.find_nearest(value, col="nom")
        if not elems:
            raise commons.UserInputError(table.__name__.lower(), f"{table.__name__} introuvable en base : {value}")
        return elems[0][0]

    async def _autocomplete(self, table: type[_Table], current: str, filtre=None) -> list[app_commands.Choice[str]]:
        return [
            app_commands.Choice(name=elem.nom, value=elem.nom) for elem in self.get_elems(table, current, filtre=filtre)
        ][:25]

    def get_elems(self, table: type[_Table], current: str, filtre=None) -> typing.Iterator[_Table]:
        for elem, _ in table.find_nearest(current, table.nom, sensi=0.25 if len(current) > 2 else 0, filtre=filtre):
            yield elem


class JoueurTransformer(app_commands.Transformer, _TableTransformerMixin):
    async def transform(self, interaction: discord.Interaction, value: str) -> Joueur:
        return await super()._transform(Joueur, value)

    async def autocomplete(self, interaction: discord.Interaction, current: str) -> list[app_commands.Choice[str]]:
        return await super()._autocomplete(Joueur, current)


class VivantTransformer(app_commands.Transformer, _TableTransformerMixin):
    async def transform(self, interaction: discord.Interaction, value: str) -> Joueur:
        joueur = await super()._transform(Joueur, value)
        if not joueur.est_vivant:
            raise commons.UserInputError("joueur", f"Eh oh, tu ne vois pas que {value} est mort ?")
        return joueur

    async def autocomplete(self, interaction: discord.Interaction, current: str) -> list[app_commands.Choice[str]]:
        return await super()._autocomplete(Joueur, current, filtre=Joueur.est_vivant)


class MortTransformer(app_commands.Transformer, _TableTransformerMixin):
    async def transform(self, interaction: discord.Interaction, value: str) -> Joueur:
        joueur = await super()._transform(Joueur, value)
        if not joueur.est_mort:
            raise commons.UserInputError("joueur", f"Eh oh, {value} est encore vivant aux dernière nouvelles !")
        return joueur

    async def autocomplete(self, interaction: discord.Interaction, current: str) -> list[app_commands.Choice[str]]:
        return await super()._autocomplete(Joueur, current, filtre=Joueur.est_mort)


class _CandidHaroTransformerMixin(_TableTransformerMixin):
    OK_SUFFIX = ""
    NOK_SUFFIX = ""

    async def _transform(self, table: type[_Table], value: str) -> _Table:
        await super()._transform(table, value.removesuffix(self.OK_SUFFIX).removesuffix(self.NOK_SUFFIX))

    async def get_choices(
        self, current: str, candid_haro_type: CandidHaroType, ok_mark: str, nok_mark: str
    ) -> list[app_commands.Choice[str]]:
        harotes = {haro.joueur for haro in CandidHaro.query.filter_by(type=candid_haro_type).all()}

        if len(current) <= 2 and harotes:
            return [app_commands.Choice(name=f"{joueur.nom}{self.OK_SUFFIX}", value=joueur.nom) for joueur in harotes][
                :25
            ]

        proches = super().get_elems(Joueur, current, filtre=Joueur.est_vivant)
        choix_proches = [
            app_commands.Choice(name=f"{joueur.nom}{self.NOK_SUFFIX}", value=joueur.nom) for joueur in proches
        ]
        choix_harotes = [
            app_commands.Choice(name=f"{joueur.nom}{self.OK_SUFFIX}", value=joueur.nom)
            for joueur in harotes
            if joueur not in proches
        ]
        return choix_proches[: 3 if choix_harotes else 25] + choix_harotes[:22]


class HaroteTransformer(VivantTransformer, _CandidHaroTransformerMixin):
    OK_SUFFIX = f" ✅ haro"
    NOK_SUFFIX = f" ⚠️ pas de haro"

    async def autocomplete(self, interaction: discord.Interaction, current: str) -> list[app_commands.Choice[str]]:
        return await super().get_choices(current, CandidHaroType.haro, "")


class CandidatTransformer(VivantTransformer, _CandidHaroTransformerMixin):
    OK_SUFFIX = f" ✅ candidat"
    NOK_SUFFIX = f" ⚠️ pas candidat"

    async def autocomplete(self, interaction: discord.Interaction, current: str) -> list[app_commands.Choice[str]]:
        return await super().get_choices(current, CandidHaroType.candidature)


class RoleTransformer(app_commands.Transformer, _TableTransformerMixin):
    async def transform(self, interaction: discord.Interaction, value: str) -> Role:
        role = await super()._transform(Role, value)
        if not role.actif:
            raise commons.UserInputError("role", "C'est quoi ce rôle ? Je connais pas")
        return role

    async def autocomplete(self, interaction: discord.Interaction, current: str) -> list[app_commands.Choice[str]]:
        return await super()._autocomplete(Role, current, filtre=Role.actif == True)


class CampTransformer(app_commands.Transformer, _TableTransformerMixin):
    async def transform(self, interaction: discord.Interaction, value: str) -> Camp:
        camp = await super()._transform(Camp, value)
        if not camp.public:
            raise commons.UserInputError("camp", "T'as vu ça où toi ?")
        return camp

    async def autocomplete(self, interaction: discord.Interaction, current: str) -> list[app_commands.Choice[str]]:
        return await super()._autocomplete(Camp, current, filtre=Camp.public == True)


class CommandTransformer(app_commands.Transformer):
    @staticmethod
    def _get_commands(interaction: discord.Interaction) -> dict[str, app_commands.Command]:
        if getattr(interaction.namespace, "mode") == "enable":
            return config.bot.tree.disabled_commands
        else:
            return config.bot.tree.enabled_commands

    async def transform(self, interaction: discord.Interaction, value: str) -> app_commands.Command:
        return self._get_commands(interaction)[value]

    async def autocomplete(self, interaction: discord.Interaction, current: str) -> list[app_commands.Choice[str]]:
        return [
            app_commands.Choice(name=name, value=name) for name in self._get_commands(interaction) if current in name
        ][:25]


class SubCommandTransformer(app_commands.Transformer):
    @staticmethod
    def _get_commands(interaction: discord.Interaction) -> dict[str, app_commands.Command]:
        return config.bot.tree.enabled_commands_and_subcommands

    async def transform(self, interaction: discord.Interaction, value: str) -> app_commands.Command:
        return self._get_commands(interaction)[value]

    async def autocomplete(self, interaction: discord.Interaction, current: str) -> list[app_commands.Choice[str]]:
        return [
            app_commands.Choice(name=name, value=name) for name in self._get_commands(interaction) if current in name
        ][:25]


# ---------------------------------------------------------------------------
# Commandes d'interaction avec les joueurs : input, boucles, confirmation...
# ---------------------------------------------------------------------------

# Commande générale, à utiliser à la place de bot.wait_for('message', ...)
async def wait_for_message(
    check: Callable[[discord.Message], bool],
    chan: discord.TextChannel,
) -> discord.Message:
    """Attend le premier message reçu rencontrant les critères demandés.

    Surcouche de :meth:`discord.Client.wait_for` permettant de réagir au mot-clé ``stop``.

    Args:
        check: Fonction validant ou non chaque message.
        chan: Le channel dans lequel le message est attendu.

    Returns:
        Le message reçu.

    Raises:
        .CommandExit: si le message est un des :obj:`.config.stop_keywords`
            (insensible à la casse), même si il respecte ``check``
    """
    stop_keywords = [kw.lower() for kw in config.stop_keywords]

    def trig_check(message: discord.Message):
        return check(message) or (message.channel == chan and message.content.lower() in stop_keywords)

    return await config.bot.wait_for("message", check=trig_check)


async def wait_for_react_added(message: discord.Message, process_text: bool = False) -> str | discord.Emoji:
    """Attend l'ajout d'une réaction sur un message.

    Args:
        message: Message où détecter les réactions.
        process_text: si ``True``, détecte aussi la réponse par message dans le même channel.

    Returns:
        L'emoji ajouté, ou le contenu du message textuel.
    """

    def react_check(react: discord.RawReactionActionEvent) -> bool:
        # Check REACT : bon message, bon emoji, et pas react du bot
        return react.message_id == message.id

    react_task = asyncio.create_task(config.bot.wait_for("raw_reaction_add", check=react_check))

    if process_text:
        # Check MESSAGE : bon channel, pas du bot, et filtre
        def message_check(mess: discord.Message) -> bool:
            return mess.channel == message.channel and mess.author != config.bot.user

    else:
        # On process DANS TOUS LES CAS, mais juste pour détecter les stop_keywords si process_text == False
        def message_check(mess: discord.Message) -> bool:
            return False

    mess_task = asyncio.create_task(wait_for_message(check=message_check, chan=message.channel))

    done, pending = await asyncio.wait([react_task, mess_task], return_when=asyncio.FIRST_COMPLETED)
    # Le bot attend ici qu'une des deux tâches aboutisse
    for task in pending:
        task.cancel()
    done_task = next(iter(done))  # done = tâche aboutie

    if done_task == react_task:  # Réaction
        return done_task.result().emoji

    # Réponse par message / STOP
    return done_task.result().content


async def sleep(chan: discord.abc.Messageable, tps: float) -> None:
    """Attend un temps donné en avertissant l'utilisateur.

    Pause l'exécution d'une commande en affichant l'indicateur *typing*
    ("*LGBot est en train d'écrire...*") sur un salon.

    Permet d'afficher plusieurs messages d'affilée en laissant le temps
    de lire, tout en indiquant que le bot n'a pas fini d'écrire.

    Args:
        chan: salon / contexte /... sur lequel attendre.
        tps: temps à attendre, en secondes.
    """
    async with chan.typing():
        await asyncio.sleep(tps)


# ---------------------------------------------------------------------------
# Utilitaires d'emojis
# ---------------------------------------------------------------------------


def montre(heure: str | None = None) -> str:
    """Renvoie l'emoji horloge le plus proche d'une heure donnée.

    Args:
        heure: heure à représenter, au format ``"XXh"`` ou ``"XXhMM"`` (défaut : heure actuelle).

    Returns:
       🕧, 🕓, 🕝...
    """
    if heure and isinstance(heure, str):
        heure, minute = heure.split("h")
        heure = int(heure) % 12
        minute = int(minute) % 60 if minute else 0
    else:
        now = datetime.datetime.now()
        heure = now.hour % 12
        minute = now.minute

    if minute >= 45:
        heure = (heure + 1) % 12

    if 15 < minute < 45:  # Demi heure
        L = [
            "\N{CLOCK FACE TWELVE-THIRTY}",
            "\N{CLOCK FACE ONE-THIRTY}",
            "\N{CLOCK FACE TWO-THIRTY}",
            "\N{CLOCK FACE THREE-THIRTY}",
            "\N{CLOCK FACE FOUR-THIRTY}",
            "\N{CLOCK FACE FIVE-THIRTY}",
            "\N{CLOCK FACE SIX-THIRTY}",
            "\N{CLOCK FACE SEVEN-THIRTY}",
            "\N{CLOCK FACE EIGHT-THIRTY}",
            "\N{CLOCK FACE NINE-THIRTY}",
            "\N{CLOCK FACE TEN-THIRTY}",
            "\N{CLOCK FACE ELEVEN-THIRTY}",
        ]
    else:  # Heure pile
        L = [
            "\N{CLOCK FACE TWELVE OCLOCK}",
            "\N{CLOCK FACE ONE OCLOCK}",
            "\N{CLOCK FACE TWO OCLOCK}",
            "\N{CLOCK FACE THREE OCLOCK}",
            "\N{CLOCK FACE FOUR OCLOCK}",
            "\N{CLOCK FACE FIVE OCLOCK}",
            "\N{CLOCK FACE SIX OCLOCK}",
            "\N{CLOCK FACE SEVEN OCLOCK}",
            "\N{CLOCK FACE EIGHT OCLOCK}",
            "\N{CLOCK FACE NINE OCLOCK}",
            "\N{CLOCK FACE TEN OCLOCK}",
            "\N{CLOCK FACE ELEVEN OCLOCK}",
        ]
    return L[heure]


def emoji_chiffre(chiffre: int, multi: bool = False) -> str:
    """Renvoie l'emoji / les emojis chiffre correspondant à un chiffre/nombre.

    Args:
        chiffre: chiffre/nombre à représenter.
        multi: si ``True``, ``chiffre`` peut être n'importe quel entier positif, dont les chiffres seront convertis
            séparément ; sinon (par défaut), ``chiffre`` doit être un entier entre ``0`` et ``10``.

    Returns:
        0️⃣, 1️⃣, 2️⃣...
    """
    if isinstance(chiffre, int) and 0 <= chiffre <= 10:
        return ["0️⃣", "1️⃣", "2️⃣", "3️⃣", "4️⃣", "5️⃣", "6️⃣", "7️⃣", "8️⃣", "9️⃣", "🔟"][chiffre]
    elif multi and str(chiffre).isdigit():
        return "".join([emoji_chiffre(int(chr)) for chr in str(chiffre)])
    else:
        raise ValueError(
            "L'argument de tools.emoji_chiffre doit être un entier entre 0 et 10 OU un entier positif avec multi=True"
        )


def super_chiffre(chiffre: int, multi: bool = False) -> str:
    """Renvoie le(s) caractère(s) exposant correspondant à un chiffre/nombre.

    Args:
        chiffre: chiffre/nombre à représenter.
        multi: si ``True``, ``chiffre`` peut être n'importe quel entier positif, dont les chiffres seront convertis
            séparément ; sinon (par défaut), ``chiffre`` doit être un entier entre ``0`` et ``9``.

    Returns:
        ⁰, ¹, ²...
    """
    if isinstance(chiffre, int) and 0 <= chiffre <= 9:
        return ["⁰", "¹", "²", "³", "⁴", "⁵", "⁶", "⁷", "⁸", "⁹"][chiffre]
    elif multi and str(chiffre).isdigit():
        return "".join([super_chiffre(int(chr)) for chr in str(chiffre)])
    else:
        raise ValueError(
            "L'argument de tools.super_chiffre doit être un entier entre 0 et 9 OU un entier positif avec multi=True"
        )


def sub_chiffre(chiffre: int, multi: bool = False) -> str:
    """Renvoie le(s) caractère(s) indice correspondant à un chiffre/nombre.

    Args:
        chiffre: chiffre/nombre à représenter.
        multi: si ``True``, ``chiffre`` peut être n'importe quel entier positif, dont les chiffres seront convertis
            séparément ; sinon (par défaut), ``chiffre`` doit être un entier entre ``0`` et ``9``.

    Returns:
        ₀, ₁, ₂...
    """
    if isinstance(chiffre, int) and 0 <= chiffre <= 9:
        return ["₀", "₁", "₂", "₃", "₄", "₅", "₆", "₇", "₈", "₉"][chiffre]
    elif multi and str(chiffre).isdigit():
        return "".join([sub_chiffre(int(c)) for c in str(chiffre)])
    else:
        raise ValueError(
            "L'argument de tools.sub_chiffre doit être un "
            "entier entre 0 et 9 OU un entier positif avec "
            "multi=True"
        )


# ---------------------------------------------------------------------------
# Utilitaires de date / temps, notamment liées aux horaires de jeu
# ---------------------------------------------------------------------------


def heure_to_time(heure: str) -> datetime.time:
    """Convertit l'écriture d'une heure en objet :class:`datetime.time`.

    Args:
        heure: heure au format ``HHh``, ``HHhMM`` ou ``HH:MM``.

    Returns:
        Le datetime correspondant.

    Raises:
        ValueError: conversion impossible (mauvais format)
    """
    try:
        if "h" in heure:
            hh, mm = heure.split("h")
        else:
            hh, mm = heure.split(":")
        return datetime.time(int(hh), int(mm) if mm else 0)
    except ValueError as exc:
        raise ValueError(f'Valeur "{heure}" non convertible ' "en temps") from exc


def time_to_heure(tps: datetime.time, sep: str = "h", force_minutes: bool = False) -> str:
    """Convertit un objet :class:`datetime.time` en heure.

    (version maison de :meth:`datetime.time.strftime`)

    Args:
        tps: temps à convertir.
        sep: séparateur heures / minutes à utiliser (défaut ``"h"``).
        force_minutes: si ``False`` (défaut), les minutes ne sont indiquées que si différentes de ``0``.

    Returns:
        La description du temps fourni (``""`` si ``tps`` est ``None``)
    """
    if tps:
        if force_minutes or tps.minute > 0:
            return f"{tps.hour}{sep}{tps.minute:02}"
        else:
            return f"{tps.hour}{sep}"
    else:
        return ""


def next_occurrence(tps: datetime.time) -> datetime.datetime:
    """Renvoie la prochaine occurrence temporelle d'une heure donnée.

    Renvoie le prochain timestamp arrivant DANS LES HORAIRES DU JEU :
    entre :func:`.tools.fin_pause` et :func:`.tools.debut_pause`.

    Args:
        tps: heure dont on veut connaître la prochaine occurrence.

    Returns:
        Le datetime du prochain timestamp.
    """
    now = datetime.datetime.now()
    jour = now.date()
    if tps < now.time():
        # Si plus tôt dans la journée que l'heure actuelle,
        # on réfléchit comme si on était demain
        jour += datetime.timedelta(days=1)

    test_dt = datetime.datetime.combine(jour, tps)
    if test_dt < debut_pause() and not en_pause():
        # Prochaine occurence avant la pause : OK
        return test_dt

    # Sinon, programmer après la pause
    finp = fin_pause()
    jour = finp.date()
    if tps < finp.time():
        # Si plus tôt dans la journée que l'heure de reprise,
        # on réfléchit comme si on était le lendemain
        jour += datetime.timedelta(days=1)

    return datetime.datetime.combine(jour, tps)


def debut_pause() -> datetime.datetime:
    """Datetime du prochain début de pause du jeu.

    Returns:
        Le timestamp correspondant au prochain vendredi 19h.
    """
    pause_time = datetime.time(hour=19)
    pause_wday = 4  # Vendredi

    now = datetime.datetime.now()
    jour = now.date()
    if pause_time < now.time():
        # Si plus tôt dans la journée que l'heure actuelle,
        # on réfléchit comme si on était demain
        jour += datetime.timedelta(days=1)

    ddays = (pause_wday - jour.weekday()) % 7
    # Jour décalé du nombre de jours avant vendredi
    pause_jour = jour + datetime.timedelta(days=ddays)
    return datetime.datetime.combine(pause_jour, pause_time)


def fin_pause() -> datetime.datetime:
    """Datetime de la prochaine fin de pause du jeu.

    Returns:
       Le timestamp correspondant au prochain dimanche 19h.
    """
    reprise_time = datetime.time(hour=19)
    reprise_wday = 6  # Dimanche

    now = datetime.datetime.now()
    jour = now.date()
    if reprise_time < now.time():
        # Si plus tôt dans la journée que l'heure actuelle,
        # on réfléchit comme si on était demain
        jour += datetime.timedelta(days=1)

    ddays = (reprise_wday - jour.weekday()) % 7
    # Jour décalé du nombre de jours avant vendredi
    reprise_jour = jour + datetime.timedelta(days=ddays)
    return datetime.datetime.combine(reprise_jour, reprise_time)


def en_pause() -> bool:
    """Détermine si le jeu est actuellement en pause hebdomadaire.

    Si il n'y a pas de pause (:func:`.fin_pause` = :func:`.debut_pause`), renvoie toujours ``False``.

    Returns:
        Si le jeu est actuellement en pause hebdomadaire.
    """
    return fin_pause() < debut_pause()


# ---------------------------------------------------------------------------
# Split et log
# ---------------------------------------------------------------------------


def smooth_split(mess: str, N: int = 1990, sep: str = "\n", rep: str = "") -> list[str]:
    """Sépare un message en une blocs moins longs qu'une limite donnée.

    Très utile pour envoyer des messages de (potentiellement) plus de 2000 caractères (limitation Discord).

    Args:
        mess: message à découper.
        N: taille maximale des messages formés (défaut ``1990``, pour avoir un peu de marge par rapport
            à la limitation, et permettre d'entourer de ``````` par exemple)
        sep: caractères où séparer préférentiellement le texte (défaut : sauts de ligne).
            Si ``mess`` contient une sous-chaîne plus longue que ``N`` ne contenant pas ``sep``,
            le message sera tronqué à la limite.
        rep: chaîne ajoutée à la fin de chaque message formé (tronqué du séparateur final) (défaut : aucune).

    Returns:
        La liste des fractions du message.
    """
    if not mess or len(mess) <= N:
        return [mess]

    mess = str(mess)
    messages_list = []
    psl = 0  # indice du Précédent Saut de Ligne
    L = len(mess)
    while psl + N < L:
        if mess.count(sep, psl, psl + N + len(sep)):
            # +len(sep) parce que si sep est à la fin, on le dégage
            i = psl + N - mess[psl : psl + N + len(sep)][::-1].find(sep)
            # un peu sombre mais vrai, tkt frère
            messages_list.append(mess[psl:i] + rep)
            psl = i + 1  # on élimine le \n
        else:
            messages_list.append(mess[psl : psl + N] + rep)
            psl += N

    if psl < L:
        messages_list.append(mess[psl:])  # ce qui reste
    return messages_list


class _UpFollower:
    def __init__(self, interaction: discord.Interaction) -> None:
        self.followup = interaction.followup
        self.channel = interaction.channel

    async def send(self, content: str | None, view: discord.ui.View | None = None, **kwargs) -> discord.Message:
        try:
            if view:
                return await self.followup.send(content, view=view, **kwargs)
            else:
                fol = self.followup
                co = fol.send(content, **kwargs)
                return await co
        except discord.HTTPException:
            if view:
                return await self.channel.send(content, view=view, **kwargs)
            else:
                return await self.channel.send(content, **kwargs)


async def _send_messages(
    messageable: discord.abc.Messageable | discord.Interaction,
    contents: list[str],
    view: discord.ui.View | None = None,
    ephemeral: bool = False,
    author: discord.Member | None = None,
    **kwargs,
) -> list[discord.Message]:
    if not contents:
        return []

    messages = []
    if isinstance(messageable, discord.Interaction):
        interaction = messageable
        # first message -> interaction reply, if possible
        if not interaction.is_expired() and not interaction.response.is_done():
            try:
                if view and len(contents) == 1:  # sole message -> display view, if applicable
                    await interaction.response.send_message(contents[0], view=view, ephemeral=ephemeral, **kwargs)
                else:
                    await interaction.response.send_message(contents[0], ephemeral=ephemeral, **kwargs)
            except discord.HTTPException:
                pass  # Problème "Unknown interaction", notamment
            else:
                contents.pop(0)
                messages.append(await interaction.original_response())

            if not contents:
                return messages

        channel = _UpFollower(interaction)
        original_channel = interaction.channel
    else:
        channel = messageable
        original_channel = channel

    if ephemeral:
        # Message éphémère : on ne peut pas l'envoyer hors contexte d'intégration -> envoi en privé à la place
        if not author:
            raise RuntimeError("Interaction context unavailable or expired, ephemeral message and no author provided")
        if author.top_role >= config.Role.mj:
            channel = config.Channel.logs
        else:
            try:
                joueur = Joueur.from_member(author)
            except ValueError:
                raise RuntimeError(
                    "Interaction context unavailable or expired, ephemeral message and author is not a player or MJ"
                ) from None
            channel = joueur.private_chan

        if channel != original_channel:
            await channel.send(
                f":warning: {author.mention} Afin que le secret de ton action reste entre nous, "
                "continuons ici :arrow_down:"
            )

    if not messages:
        first_content = contents.pop(0)
        if view and not contents:  # last message -> display view, if applicable
            messages.append(await channel.send(first_content, view=view, **kwargs))
        else:
            messages.append(await channel.send(first_content, **kwargs))
        if not contents:
            return messages

    *main_contents, last_content = contents
    for content in main_contents:
        messages.append(await channel.send(content))

    # last message -> display view, if applicable
    messages.append(await channel.send(last_content, view=view))

    return messages


async def send_blocs(
    messageable: discord.abc.Messageable | discord.Interaction,
    mess: str,
    *,
    N: int = 1990,
    sep: str = "\n",
    rep: str = "",
    code: bool = False,
    prefix: str = "",
    langage: str = "",
    **kwargs,
) -> list[discord.Message]:
    """Envoie un message en le coupant en blocs si nécessaire.

    Surcouche de :func:`.smooth_split` envoyant directement les messages formés.

    Args:
        messageable: objet où envoyer le message ou interaction.
        mess: message à envoyer
        N, sep, rep: passé à :func:`.smooth_split`.
        prefix: texte optionnel à mettre hors des code blocs, au début du premier message.
        language: voir :func:`.code_bloc`.

    Returns:
        La liste des messages envoyés.
    """
    mess = str(mess)
    if prefix:
        prefix = prefix.rstrip() + "\n"

    blocs = smooth_split(prefix + mess, N=N, sep=sep, rep=rep)

    if code and mess:
        if prefix:
            blocs[0] = blocs[0][len(prefix) :]
        blocs = [code_bloc(bloc, langage=langage) for bloc in blocs]
        if prefix:
            blocs[0] = prefix + blocs[0]

    return await _send_messages(messageable, blocs, **kwargs)


async def log(
    message: str,
    *,
    code: bool = False,
    N: int = 1990,
    sep: str = "\n",
    rep: str = "",
    prefix: str = "",
    langage: str = "",
) -> list[discord.Message]:
    """Envoie un message dans le channel :attr:`config.Channel.logs`.

    Surcouche de :func:`.send_blocs` / :func:`.send_code_blocs`.

    Args:
        message: message à log.
        code: si ``True``, log sous forme de bloc(s) de code (défaut ``False``).
        N, sep, rep: passé à :func:`.send_blocs` / :func:`.send_code_blocs`.
        prefix: voir :func:`.send_blocs`, simplement ajouté avant ``message`` si ``code`` vaut ``False``.
        language: *identique à* :func:`.send_code_blocs`, ignoré si `code` vaut ``False``.

    Returns:
        La liste des messages envoyés.
    """
    return await send_blocs(
        config.Channel.logs, message, N=N, sep=sep, rep=rep, code=code, prefix=prefix, langage=langage
    )


# ---------------------------------------------------------------------------
# Autres fonctions diverses
# ---------------------------------------------------------------------------


def remove_accents(text: str) -> str:
    """Enlève les accents d'un chaîne, mais conserve les caractères spéciaux.

    Version plus douce de ``unidecode.unidecode``, conservant notamment les emojis, ...

    Args:
        text: chaîne à désaccentuer.

    Returns:
        La chaîne désaccentuée.
    """
    p = re.compile("([À-ʲΆ-ת])")
    # Abracadabrax, c'est moche mais ça marche (source : tkt frère)
    return p.sub(lambda c: unidecode.unidecode(c.group()), text)


# Évaluation d'accolades
def eval_accols(rep: str, globals_: dict[str, Any] = None, locals_: dict[str, Any] = None, debug: bool = False) -> str:
    """Replace chaque bloc entouré par des ``{}`` par leur évaluation Python.

    Args:
        globals_: variables globales du contexte d'évaluation (passé à :func:`eval`).
        locals_: variables locales du contexte d'évaluation (passé à :func:`eval`).
        debug: si ``True``, insère le message d'erreur (type et texte de l'exception) dans le message
            à l'endroit où une exception est levée durant l'évaluation (défaut ``False``).

    Penser à passer les :func:`globals` et :func:`locals` si besoin.
    Généralement, il faut passer :func:`locals` qui contient ``ctx``, etc...
    mais pas :func:`globals` si on veut bénéficier de tous les modules importés dans ``tools.py``.
    """
    if globals_ is None:
        globals_ = globals()
        globals_.update(tools=__import__(__name__, fromlist=("tools")))
    if locals_ is None:
        locals_ = globals_

    if "{" not in rep:  # Si pas d'expressions, on renvoie direct
        return rep

    evrep = ""  # Réponse évaluée
    expr = ""  # Expression à évaluer
    noc = 0  # Nombre de { non appariés
    for car in rep:
        if car == "{":
            if noc:  # Expression en cours :
                expr += car  # on garde le {
            noc += 1
        elif car == "}":
            noc -= 1
            if noc:  # idem
                expr += car
            else:  # Fin d'une expression
                try:  # On essaie d'évaluer la chaîne
                    evrep += str(eval(expr, globals_, locals_))
                except Exception as e:
                    # Si erreur, on laisse {expr} non évaluée
                    evrep += "{" + expr + "}"
                    if debug:
                        evrep += code(f"->!!! {e} !!!")
                expr = ""
        elif noc:  # Expression en cours
            expr += car
        else:  # Pas d'expression en cours
            evrep += car
    if noc:  # Si expression jamais finie (nombre impair de {)
        evrep += "{" + expr
    return evrep


async def multicateg(base_name: str) -> discord.CategoryChannel:
    """Permet de gérer des groupes de catégories (plus de 50 salons).

    Renvoie la première catégorie pouvant accueillir un nouveau
    salon ; en crée une nouvelle si besoin.

    Args:
        base_name: le nom du groupe de catégorie (nom de la première catégorie, puis sera suivi de 2, 3...)

    Warning:
        Une catégorie appelée ``base_name`` doit exister au préalable dans le serveur (:attr:`config.guild`).

    Returns:
       La catégorie libre.
    """
    categ = channel(base_name)
    nb = 1
    while len(categ.channels) >= 50:
        # Limitation Discord : 50 channels par catégorie
        nb += 1
        next_name = f"{base_name} {nb}"
        next_categ = channel(next_name, must_be_found=False)
        if not next_categ:
            next_categ = await categ.clone(name=next_name)
        categ = next_categ

    return categ


def in_multicateg(categ: discord.CategoryChannel, base_name: str) -> bool:
    """Détecte si une catégorie appartient à un groupe de catégories.

    Args:
        categ: la catégorie à tester.
        base_name: le nom de base du groupe de catégories (voir :func:`.multicateg`).

    Returns:
        Si la catégorie fait partie du groupe.
    """
    stripped_name = categ.name.rstrip(string.digits + " ")
    return stripped_name == base_name


# ---------------------------------------------------------------------------
# Utilitaires de formatage de texte
# ---------------------------------------------------------------------------


def bold(text: str) -> str:
    """Formate une chaîne comme texte en **gras** dans Discord.

    Args:
        text: chaîne à formater.

    Returns:
        La chaîne formatée.
    """
    return f"**{text}**"


def ital(text: str) -> str:
    """Formate une chaîne comme texte en *italique* dans Discord.

    Args:
        text: chaîne à formater.

    Returns:
        La chaîne formatée.
    """
    return f"*{text}*"


def soul(text: str) -> str:
    """Formate une chaîne comme texte souligné dans Discord.

    Args:
        text: chaîne à formater.

    Returns:
        La chaîne formatée.
    """
    return f"__{text}__"


def strike(text: str) -> str:
    """Formate une chaîne comme texte barré dans Discord.

    Args:
        text: chaîne à formater.

    Returns:
        La chaîne formatée.
    """
    return f"~~{text}~~"


def code(text: str) -> str:
    """Formate une chaîne comme ``code`` (inline) dans Discord.

    Args:
        text: chaîne à formater.

    Returns:
        La chaîne formatée.
    """
    return f"`{text}`"


def code_bloc(text: str, langage: str = "") -> str:
    """Formate une chaîne comme un bloc de code dans Discord.

    Args:
        text: chaîne à formater.
        langage: langage du code, pour coloration syntaxique.

    Langages supportés (non exhaustif ?) : ``asciidoc``, ``autohotkey``,
    ``bash``, ``coffeescript``, ``cpp`` (C++), ``cs`` (C#), ``css``,
    ``diff``, ``fix``, ``glsl``, ``ini``, ``json``, ``md``, (markdown),
    ``ml``, ``prolog``, ``py``, ``tex``, ``xl``, ``xml``

    Returns:
        La chaîne formatée.
    """
    return f"```{langage}\n{text}```"


def quote(text: str) -> str:
    """Formate une chaîne comme citation (inline) dans Discord.

    Args:
        text: chaîne à formater.

    Returns:
        La chaîne formatée.
    """
    return f"> {text}"


def quote_bloc(text: str) -> str:
    """Formate une chaîne comme bloc de citation (multiline) dans Discord.

    Args:
        text: chaîne à formater.

    Returns:
        La chaîne formatée.
    """
    return f">>> {text}"


def spoiler(text: str) -> str:
    """Formate une chaîne comme spoiler (cliquer pour afficher) dans Discord.

    Args:
        text: chaîne à formater.

    Returns:
        La chaîne formatée.
    """
    return f"||{text}||"



"""lg-rez / features / Tâches planifiées

Planification, liste, annulation, exécution de tâches planifiées

"""

import datetime

import discord
from discord import app_commands

from lgrez import commons
from lgrez.blocs import tools
from lgrez.bdd import Tache, Action
from lgrez.blocs.journey import DiscordJourney, journey_command
from lgrez.features import communication


DESCRIPTION = """Commandes de planification, exécution, annulation de tâches"""


class TimestampTransformer(app_commands.Transformer):
    FAST_CHOICES = {
        "Demain 7h": "7:01",
        "Demain 9h": "9:01",
    }

    async def transform(self, interaction: discord.Interaction, value: str) -> datetime.datetime:
        if value in self.FAST_CHOICES:
            value = self.FAST_CHOICES[value]

        now = datetime.datetime.now()

        for format in ["%d/%m/%Y %H:%M:%S", "%d/%m/%Y %H:%M", "%d/%m/%Y %Hh%M", "%d/%m/%Y %Hh"]:
            try:
                return datetime.datetime.strptime(value, format)
            except ValueError:
                continue

        for format in ["%d/%m %H:%M:%S", "%d/%m %H:%M", "%d/%m %Hh%M", "%d/%m %Hh"]:
            try:
                return datetime.datetime.strptime(value, format).replace(year=now.year)
            except ValueError:
                continue

        for format in ["%H:%M:%S", "%H:%M", "%Hh%M", "%Hh"]:
            try:
                return datetime.datetime.strptime(value, format).replace(year=now.year, month=now.month, day=now.day)
            except ValueError:
                continue

        raise commons.UserInputError("quand", "Format invalide (doit être [DD/MM[/YYYY]] HH:MM[:SS]")

    async def autocomplete(self, interaction: discord.Interaction, current: str) -> list[app_commands.Choice[str]]:
        if not current:
            return [app_commands.Choice(name=name, value=value) for name, value in self.FAST_CHOICES.items()]
        return []


@app_commands.command()
@tools.mjs_only
@journey_command
async def taches(journey: DiscordJourney):
    """Liste les tâches actuellement planifiées (COMMANDE MJ)

    Affiche les commandes en attente d'exécution et le timestamp d'exécution associé.
    Lorsque la tâche est liée à une action, affiche le nom de l'action et du joueur concerné.
    """
    lst: list[Tache] = Tache.query.order_by(Tache.timestamp).all()
    rep = ""
    for tache in lst:
        rep += f"\n{str(tache.id).ljust(5)} {tache.timestamp:%d/%m/%Y %H:%M:%S}    {tache.description.ljust(25)} "
        if action := tache.action:
            rep += f"{action.base.slug.ljust(20)} {action.joueur.nom}"

    if rep:
        prefix = "Tâches en attente :"
        mess = (
            "ID    Timestamp              Commande                  Action               Joueur\n"
            f"{'-' * 105}{rep}\n\n"
            "Utilisez /cancel <ID> pour annuler une tâche."
        )
    else:
        prefix = "Aucune tâche en attente."
        mess = ""

    await journey.send(mess, code=True, prefix=prefix)


planif = app_commands.Group(name="planif", description="Planification de commandes")


async def _planif(journey: DiscordJourney, quand: datetime.datetime, command: app_commands.Command, **parameters):
    mess = ""
    if quand < datetime.datetime.now():
        quand = datetime.datetime.combine(datetime.date.today() + datetime.timedelta(days=1), quand.time())
        mess = ":warning: Date dans le passé, décalée à demain\n"

    tache = await planif_command(quand, command, **parameters)

    await journey.send(
        mess
        + f":arrow_forward: Commande `{tache.description}` planifiée pour le {quand:%d/%m/%Y} à {quand:%H:%M:%S}.\n"
        f"`/cancel {tache.id}` pour annuler."
    )


async def planif_command(timestamp: datetime.datetime, command: app_commands.Command, **parameters) -> Tache:
    action = None
    # ID de l'action associée à la tâche le cas échéant
    if command.name == "action" and (id := parameters.get("id")):  # open / close / remind
        action = Action.query.get(int(id))

    tache = Tache(timestamp=timestamp, commande=command.qualified_name, parameters=parameters, action=action)
    tache.add()  # Planifie la tâche
    return tache


@planif.command()
@tools.mjs_only
@journey_command
async def post(
    journey: DiscordJourney,
    *,
    quand: app_commands.Transform[datetime.datetime, TimestampTransformer],
    chan: discord.TextChannel,
    message: str,
):
    """Planifie l'envoi d'un message dans un salon (raccourci pour /planif command, COMMANDE MJ)

    Args:
        quand: Quand planifier la commande (date optionnelle, défaut aujourd’hui, année / minutes / secondes aussi).
        chan: Salon ou poster le message.
        message: Message à envoyer (utiliser "\n" pour un saut de ligne).

    Si la date spécifiée est dans le passé, la commande est planifiée pour le lendemain.
    """
    await _planif(journey, quand, communication.post, chan=chan, message=message)


@planif.command()
@tools.mjs_only
@journey_command
async def command(
    journey: DiscordJourney,
    *,
    quand: app_commands.Transform[datetime.datetime, TimestampTransformer],
    command: app_commands.Transform[app_commands.Command, tools.SubCommandTransformer],
):
    """Planifie l'exécution d'une commande quelconque (COMMANDE MJ)

    Args:
        quand: Quand planifier la commande (date optionnelle, défaut aujourd’hui, année / minutes / secondes aussi).
        command: Commande à lancer.

    Ne pas planifier de commandes avec confirmation / modale !

    Si la date spécifiée est dans le passé, la commande est planifiée pour le lendemain.
    """
    _callback, parameters = await journey.command_parameters_modal(command, title_prefix=f"/planif {quand}")
    await _planif(journey, quand, command, **parameters)


@app_commands.command()
@tools.mjs_only
@journey_command
async def cancel(journey: DiscordJourney, *, id: int):
    """Annule une tâche planifiée (COMMANDE MJ)

    Args:
        id: ID de la tâche à annuler (voir /taches).

    Utiliser ``/taches`` pour voir la liste des IDs.
    """
    if tache := Tache.query.get(int(id)):
        tache.delete()
        await journey.send(f"Tâche annulée :\n```/planif command {tache.timestamp:%d/%m/%Y %X}\n{tache.description}```")
    else:
        await journey.send(f":x: Tâche #{id} introuvable.")



"""lg-rez / blocs / Interfaçage Google Sheets

Connection, récupération de classeurs, modifications
(implémentation de https://pypi.org/project/gspread)
"""

from __future__ import annotations

import enum
import functools
import json
import typing

import gspread
import gspread_asyncio
import requests
from oauth2client import service_account
from googleapiclient.discovery import build

from lgrez import bdd
from lgrez.blocs import env


WorksheetNotFound = gspread.exceptions.WorksheetNotFound
ConnectionError = requests.exceptions.ConnectionError


class Modif:
    """Modification à appliquer à un Google Sheet.

    Attributes:
        row: Numéro de la ligne (0 = ligne 1)
        column: Numéro de la colonne (0 = colonne A)
        val: Nouvelle valeur
    """

    def __init__(self, row: int, column: int, val: typing.Any) -> None:
        """Initializes self."""
        self.row = row
        self.column = column
        self.val = val

    def __repr__(self) -> str:
        """Returns repr(self)"""
        return f"<gsheets.Modif: ({self.row}, {self.column}) = {self.val!r}>"

    def __eq__(self, other: Modif) -> bool:
        """Returns self == other"""
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.row == other.row and self.column == other.column and self.val == other.val

    def __hash__(self) -> int:
        return hash((self.row, self.column, self.val))


@functools.cache
def _get_creds(*scopes: str) -> service_account.ServiceAccountCredentials:
    # use creds to create a client to interact with the Google Drive API
    LGREZ_GCP_CREDENTIALS = env.load("LGREZ_GCP_CREDENTIALS")
    return service_account.ServiceAccountCredentials.from_json_keyfile_dict(
        json.loads(LGREZ_GCP_CREDENTIALS), list(scopes)
    )


async def connect(key: str) -> gspread_asyncio.AsyncioGspreadWorksheet:
    """Charge les credentials GSheets et renvoie le classeur demandé.

    Nécessite la variable d'environment ``LGREZ_GCP_CREDENTIALS``.

    Args:
        key: ID du classeur à charger (25 caractères)

    Returns:
        Le classeur correspondant à l'ID demandé.
    """
    scope = "https://spreadsheets.google.com/feeds"
    manager = gspread_asyncio.AsyncioGspreadClientManager(lambda: _get_creds(scope))
    client = await manager.authorize()

    # Open the workbook
    workbook = await client.open_by_key(key)

    return workbook


async def update(
    sheet: gspread_asyncio.AsyncioGspreadWorksheet,
    *modifs: Modif,
) -> None:
    """Met à jour une feuille GSheets avec les modifications demandées.

    Args:
        sheet: La feuille à modifier.
        *modifs: Les modification(s) à apporter.

    Le type de la nouvelle valeur sera interprété par ``gspread`` pour
    donner le type GSheets adéquat à la cellule (texte, numérique,
    temporel...)

    Les entiers trop grands pour être stockés sans perte de précision
    (IDs des joueurs par exemple) sont convertis en :class:`str`. Les
    ``None`` sont convertis en ``''``. Les membres d':class:`~enum.Enum`
    sont stockés par leur **nom**.
    """
    # Bordures de la zone à modifier
    lm = max([modif.row for modif in modifs])
    cm = max([modif.column for modif in modifs])

    # Récupère toutes les valeurs sous forme de cellules gspread
    cells = await sheet.range(1, 1, lm + 1, cm + 1)
    # gspread indexe à partir de 1 (comme les gsheets)

    cells_to_update = []
    for modif in modifs:
        # On récupère l'objet Cell correspondant aux coords à modifier
        cell = next(cell for cell in cells if cell.col == modif.column + 1 and cell.row == modif.row + 1)

        val = modif.val

        # Transformation objets complexes
        if isinstance(val, enum.Enum):
            # Enums : stocker le nom
            val = val.name
        elif isinstance(modif.val, bdd.base.TableBase):
            # Instances : stocker la clé primaire
            val = val.primary_key

        # Adaptation types de base
        if isinstance(val, int) and val > 10**14:
            # Entiers trop grands pour être stockés sans perte de
            # précision (IDs des joueurs par ex.): passage en str
            cell.value = str(val)
        elif val is None:
            cell.value = ""
        else:
            cell.value = val

        cells_to_update.append(cell)

    await sheet.update_cells(cells_to_update)


def a_to_index(column: str) -> int:
    """Utilitaire : convertit une colonne ("A", "B"...) en indice.

    Args:
        column: nom de la colonne. Doit être composé de caractères
            dans [a-z, A-Z] uniquement.

    Returns:
        L'indice de la colonne, **indexé à partir de 0** (cellules
        considérées comme une liste de liste).

    Raises:
        gspread.exceptions.IncorrectCellLabel: valeur incorrecte.
    """
    a1 = column + "1"
    row, col = gspread.utils.a1_to_rowcol(a1)
    return col - 1  # a1_to_rowcol indexe à partir de 1


def get_doc_content(doc_id: str) -> list[tuple[str, dict]]:
    """Récupère le contenu d'un document Google Docs.

    Transforme le document pour ne garder que les fragments de
    texte et leur mise en forme. Les paragraphes de listes à
    points sont transformés en ajoutant un fragment ``    -  ``
    à l'endroit adéquat ; les autres options de mise en forme des
    paragraphes et les autres objets GDocs sont ignorés.

    Args:
        doc_id: ID du document à récupérer (doit être public
            ou dans le Drive partagé avec le compte de service).

    Returns:
        Les différents fragments de texte du document et leur formattage
        (référence : https://developers.google.com/docs/api/reference/rest/v1/documents#TextStyle)

    Raises:
        googleapiclient.errors.HttpError: ID incorrect ou document non
            accessible.
    """
    scope = "https://www.googleapis.com/auth/documents.readonly"
    service = build("docs", "v1", credentials=_get_creds(scope))

    # Retrieve the documents contents from the Docs service.
    document = service.documents().get(documentId=doc_id).execute()

    content = []
    blocs = document["body"]["content"]
    for bloc in blocs:
        paragraph = bloc.get("paragraph")
        if not paragraph:
            continue

        bullet = paragraph.get("bullet")
        if bullet:
            content.append(("    -  ", bullet["textStyle"]))

        elements = paragraph["elements"]
        for element in elements:
            run = element.get("textRun")
            if run:  # bout de texte
                content.append((run["content"], run["textStyle"]))

    return content


def get_files_in_folder(folder_id: str) -> list[dict[str, str]]:
    """Récupère le contenu binaire d'un fichier Google Drive.
    Args:
        folder_id: ID du fichier à récupérer (doit être public ou
            dans le Drive partagé avec le compte de service).
    Returns:
        A list of, for each file, a directory with ``"file_id"``,
        ``name`` (with extension) and ``extension`` (without dot) data.
    Raises:
        googleapiclient.errors.HttpError: ID incorrect ou dossier non
            accessible.
        RuntimeError: Autre erreur.
    """
    scope = "https://www.googleapis.com/auth/drive.readonly"
    service = build("drive", "v3", credentials=_get_creds(scope))
    data = (
        service.files()
        .list(
            corpora="user",
            q=f"'{folder_id}' in parents",
            fields="files(id, fileExtension, name)",
        )
        .execute()
    )
    if not data:
        raise RuntimeError(f"Unable to get Drive folder info '{folder_id}'")
    files = [
        {
            "file_id": file_data.get("id", ""),
            "name": file_data.get("name", ""),
            "extension": file_data.get("fileExtension", ""),
        }
        for file_data in data.get("files", [])
    ]
    return files


def download_file(file_id: str) -> bytes:
    """Récupère le contenu binaire d'un fichier Google Drive.
    Args:
        file_id: ID du fichier à récupérer (doit être public ou
            dans le Drive partagé avec le compte de service).
    Returns:
        Le contenu binaire du fichier.
    Raises:
        googleapiclient.errors.HttpError: ID incorrect ou document non
            accessible.
        RuntimeError: Autre erreur.
    """
    scope = "https://www.googleapis.com/auth/drive.readonly"
    service = build("drive", "v3", credentials=_get_creds(scope))
    data = service.files().get_media(fileId=file_id).execute()
    if not data:
        raise RuntimeError(f"Unable to download Drive file '{file_id}'")
    return data



"""lg-rez / blocs / Console d'administration (avancé)

Permet de contrôler le bot depuis la console le faisant tourner.

"""

import os
import sys
import asyncio

import aioconsole
import asyncode
import discord

from lgrez import config


class CaptureLogs:
    """Context manager: capture messages posted to :attr:`.config.Channel.logs`.

    Only capture messages posted by the bot itself: these messages
    are not sent at all to Discord.
    """

    def __init__(self) -> None:
        """Initialize self."""
        self._chan = config.Channel.logs
        self._calls = []

        async def _cor(text: str, *args, **kwargs):
            self._calls.append(text)

        class _ProxyChannel(discord.TextChannel):
            def __init__(self) -> None:
                # We inherit TextChannel for type checks, but we
                # do NOT want to create a new TextChannel object
                pass

            def __getattribute__(slf, attr):
                if attr == "send":
                    return _cor
                return getattr(self._chan, attr)

        self.proxy = _ProxyChannel()

    def __enter__(self):
        """Enter the context: start capturing."""
        config.Channel.logs = self.proxy
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context: stop capturing."""
        config.Channel.logs = self._chan

    @property
    def text(self) -> str:
        """The messages sent in the channel, joined with newlines."""
        return "\n".join(self._calls)


class AdminConsole(asyncode.AsyncInteractiveConsole):
    """Python console to control LGBot from the running shell.

    (subclass of :class:`asyncode.AsyncInteractiveConsole`)

    Args:
        locals, filename: see :class:`asyncode.AsyncInteractiveConsole`.
    """

    def __init__(
        self,
        locals: dict = {},
        filename: str = "__console__",
    ) -> None:
        """Initialize self"""
        locals["_console"] = self
        super().__init__(locals, filename)

    async def interact(
        self,
        banner: str | None = None,
        exitmsg: str | None = "Byebye!",
    ) -> None:
        """Launches the console.

        (see :meth:`asyncode.AsyncInteractiveConsole.interact`)
        """
        await aioconsole.aprint("\nLaunching interactive console...")

        if banner is None:
            base = (
                'Type "help", "copyright", "credits" or "license" for '
                'more information.\nasync REPL: use "await" directly '
                "instead of asyncio.run()."
            )
            warn = "WARNING: Sending KeyboardInterrupt (^C) will stop the whole LGBot."
            banner = f"Python {sys.version} on {sys.platform}\n{base}\n{warn}"

        await super().interact(banner, exitmsg)

    async def write(self, data: str) -> None:
        """Method called on each print / repr / traceback...

        Relies on :meth:`aioconsole.aprint`.

        Args:
            data: text to display.
        """
        if not hasattr(sys, "ps1"):
            sys.ps1 = ">>> "
        if not hasattr(sys, "ps2"):
            sys.ps2 = "... "

        await aioconsole.aprint(data, end="")

    async def raw_input(self, prompt: str = "") -> str:
        """Method called when the console waits for next instruction.

        Relies on :meth:`aioconsole.ainput`.

        Args:
            prompt: optional text to send before waiting for input.

        Returns:
            The new instruction to compute.
        """
        if prompt:
            await self.write(prompt)

        return await aioconsole.ainput()


async def run_admin_console(locals: dict) -> None:
    """Launch the Admin console and restart it if crashing.

    Args:
        locals: the objects accessible from the console.
    """
    if os.isatty(sys.stdin.fileno()):  # Lancement depuis un terminal
        await asyncio.sleep(2)
        cons = AdminConsole(locals=locals)
        await cons.interact()
        await aioconsole.aprint("Admin console closed (bot still runs).")


async def execute_command(entree: str) -> str:
    """Execute a command.

    Create a fake message in the #logs channel by the server owner,
    and process it while capturing the output in the channel.

    Args:
        entree: The command to execute (must begin with the
            command prefix)

    Returns:
        The messages get in response to the command execution,
        joined with newlines.
    """
    message = (await config.Channel.logs.history(limit=1).flatten())[0]
    # On a besoin de récupérer un message, ici le dernier de #logs
    message.author = config.guild.owner
    message.content = entree
    ctx = await config.bot.get_context(message)

    _calls = []

    async def _cor(text, *args, **kwargs):
        _calls.append(text)

    ctx.send = _cor
    ctx.reply = _cor

    with CaptureLogs() as cc:
        await config.bot.invoke(ctx)

    text = "\n".join(_calls)
    if cc.text:
        text += f"\n[LOGGED] > {cc.text}"
    text = text.replace('"""', '\\"\\"\\"')
    return f'print("""{text}""")'
